/***************************************************************************
* GEM -- electrostatics calculations and visualization                     *
* Copyright (C) 2006 John C. Gordon, 2021 Dan Folescu                      *
*                                                                          *
* This program is free software; you can redistribute it and/or modify     *
* it under the terms of the GNU General Public License as published by     *
* the Free Software Foundation; either version 2 of the License, or        *
* (at your option) any later version.                                      *
*                                                                          *
* This program is distributed in the hope that it will be useful,          *
* but WITHOUT ANY WARRANTY; without even the implied warranty of           *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
* GNU General Public License for more details.                             *
*                                                                          *
* You should have received a copy of the GNU General Public License along  *
* with this program; if not, write to the Free Software Foundation, Inc.,  *
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.              *
***************************************************************************/

#include "file_io.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/wait.h>

//Helper function to generate the configuration files with the given parameters
//pR	-	probe radius(default 1.4)
//gD	-	grid density(inverse grid cube side length[Angstrom], default 2)
//sIGC	-	self intersection grid coefficient(default 1.5)
//mP	-	max probes(default 100)
static void generate_configuration(char* configuration_file_name, char* xyzr_filename, double pR, double gD, double sIGC, int mP);

//Helper function to generate the configuration files with the given parameters
//pR	-	probe radius
//gD	-	grid density
//sIGC	-	self intersection grid coefficient
//mP	-	max probes
static void generate_configuration(char* configuration_file_name, char* xyzr_filename, double pR, double gD, double sIGC, int mP){
	//Copys default configuration parameters to a new config file
	FILE *gC = fopen(configuration_file_name, "w");

	fprintf(gC, "XYZR_FileName=%s\n", xyzr_filename);	//input file to nanoshaper - of xyzr type
	fprintf (gC, "Skin_Surface_Parameter=.45\n");	//CGAL skin parameter for surface. .45 is experimentally optimal by NS authors
	fprintf (gC, "Surface=ses\n");					//Specifies the surface to be of ses/ConnolySurface type
	fprintf (gC, "Vertex_Atom_Info=true\n");	//necessary for nearest atom information
	fprintf (gC, "Skin_Fast_Projection=true\n");	//Faster projection at the expense of accuracy. Test on/off.
	fprintf (gC, "Accurate_Triangulation=true\n");	//Analytic triangulation where possible. Probably most accurate.
	fprintf (gC, "Triangulation=true\n"); 			//specifies triangulation of the surface - Required.
	fprintf (gC, "Smooth_Mesh=false\n"); 			//more visually pleasing, but affects representation parameters. Test for accuracy
	fprintf (gC, "Compute_Vertex_Normals=true\n");	//necessary for output of nanoshaper format(.face/.vertex) files
	fprintf (gC, "Save_Mesh_MSMS_Format=true\n");	//option to save in MSMS format
	fprintf (gC, "Operative_Mode=normal\n");		//specifies a normal triangulation run
	fprintf (gC, "Grid_perfil=50\n");				//specifies the percentage of the surface to be occupied by the designated grid. Definitely needs testing for accuracy, runtime, and memory usage
	//adding input parameters to generated configuration file
	fprintf (gC, "Probe_Radius=%f\n", pR);
	fprintf (gC, "Grid_scale=%f\n", gD);
	fprintf (gC, "Self_Intersections_Grid_Coefficient=%f\n", sIGC);
	fprintf (gC, "Max_Probes_Self_Intersections=%d\n", mP);

	fclose(gC);
}

int run_nanoshaper(physical_structure *phys_struct, model_parameters *md_param)
{
  char *nanoshaper_exec_path;
  char xyzr_filename[1024];
  char line[1024];

  //path to nanoshaper executable
  nanoshaper_exec_path = getenv("NANOSHAPER_EXECUTABLE_PATH");

  //filename of soon-to-be xyzr file
  sprintf(xyzr_filename, "%s.xyzr", phys_struct->mol_name);

  write_xyzr(xyzr_filename, phys_struct);
  
  if(md_param->configuration_file_name == NULL){
    md_param->configuration_file_name = calloc(strlen("generated.conf")+3,sizeof(char));
    strcpy(md_param->configuration_file_name,"generated.conf");
	  generate_configuration(md_param->configuration_file_name, xyzr_filename, md_param->water_probe_radius, md_param->triangulation_density, 100, 100);

	fprintf(stdout, "Running NanoShaper with given parameters\n");
  }

  int nano_status = -1;

  int val = fork();

	//In child
	if(val == 0){
		char* arg[] = {nanoshaper_exec_path, md_param->configuration_file_name, NULL};
		int nano_out = open("nano.out", O_WRONLY | O_TRUNC | O_CREAT, S_IRUSR | S_IRGRP | S_IWGRP | S_IWUSR);
		dup2(nano_out, 1);
		execvp(nanoshaper_exec_path, arg);
		exit(0);
	}
	//In parent
	else{
		//child's return status
		int status;
		//parent waits for child to die
		pid_t waited = waitpid(val, &status, 0);
		//if there was fork-exec system error
		if(waited == -1){
			fprintf(stderr, "There was an error when forking");
			nano_status = -1;
		}
		//if the child exited "normally"
		if(WIFEXITED(status)){
			//check what its last words were
			status = WEXITSTATUS(status);
			//if the child had no regrets, status == 0
			if(!status){
				nano_status = 0; //nanoshaper exited successfully
			}
    }
		//Had a system error, and should notify via stderr
		else{
			fprintf(stderr, "System error!");
			nano_status = -2;
		}
    // Rename output files to maintain programatic consistency
    if(!nano_status){
      sprintf(line, "%s.vert", phys_struct->mol_name);
      rename("triangulatedSurf.vert", line);
      // clear line between renames
      memset(line, 0 ,sizeof(line));
      sprintf(line, "%s.face", phys_struct->mol_name);
      rename("triangulatedSurf.face", line);
      return 0;
    }
    return -1;
  }
}
